@@ -59,7 +59,7 @@
 //! ```
 //!
 //! This pattern isn't too hard for a typical data structure to implement.
-//! However, the transcoder faces a couple of challenges:
+//! However, the transcoder faces a few challenges:
 //!
 //! 1. Unlike with a typical data structure, `serialize` can't know which
 //!    `serializer` method to call until it asks the deserializer to drive a new
@@ -71,38 +71,56 @@
 //!    return an `S::Error`, which it can only construct with a string-like
 //!    value via the [`ser::Error::custom`] function.
 //!
+//! 3. When `serialize` returns an error, its caller within the transcoder does
+//!    not know enough about `S::Error` to distinguish a legitimate serializer
+//!    error from a `custom` error generated by the transcoder itself. Without
+//!    this information, the transcoder cannot reliably report which side of a
+//!    transcode operation caused it to fail.
+//!
 //! In general, all of the transcoder's trait implementations face similar
 //! challenges, with the details varying based on which part of the process
 //! they're involved with. As such, these implementations all center around a
-//! common [`Machine`] type. A machine initially owns an unused (de)serializer
-//! for a trait method to take, and subsequently provides a writable slot for
-//! any error value that the method can't return directly.
+//! common [`Transcoder`] type. A `Transcoder` initially owns an unused parent
+//! (de)serializer for a trait method to take, provides storage for any error
+//! value that the method can't return directly, and tracks the original source
+//! of an error to hide synthetic errors from the transcoder's final result.
 //!
-//! A typical implementation of a Serde trait method on `Machine` will:
+//! A typical implementation of a Serde trait method on `Transcoder` will:
 //!
-//! 1. Take ownership of the (de)serializer held by the machine, replacing the
-//!    machine state with an empty error slot. (Subsequent attempts to do this
-//!    will panic.)
+//! 1. Take ownership of the parent (de)serializer held by the transcoder.
+//!    Subsequent attempts to do this will panic.
 //!
 //! 2. Call the appropriate (de)serializer method for the next step of the
 //!    transcode. It might just need to serialize a simple value, or it might
-//!    need to pass the callee a reference to a newly constructed "inner"
-//!    machine, through which the callee can invoke other Serde trait methods.
+//!    need to construct a new "inner" transcoder to allow the callee to invoke
+//!    other Serde trait methods.
 //!
-//! 3. Stash any error from that call in the machine's error slot if it can't be
-//!    returned directly. Then, return an error of the appropriate type, either
-//!    by constructing one with a generic message, or by extracting one from an
-//!    inner machine that was captured further down the call stack.
+//! 3. Capture any state from that call that the method can't return directly,
+//!    including the call's true error value and the appropriate error source
+//!    information.
 //!
-//! This plumbing of error values through the call stack ensures that a failed
-//! transcode follows the same error paths that the serializer and deserializer
-//! would normally traverse when aborting a typical (de)serialize operation in
-//! the middle of a value. An early implementation of the xt transcoder simply
-//! attempted to make its `Visitor` yield `Result<S::Ok, S::Error>` instead of
-//! `S::Ok`. However, at least one deserializer was found to panic when the
-//! transcoder attempted to return a serializer error through the deserializer's
-//! success path (by returning `Ok(Err(_))` from visitor methods) without fully
-//! consuming the deserializer's current input value.
+//! 4. Return an error of the method's return type, either by constructing one
+//!    with a generic message or extracting one from an inner transcoder.
+//!
+//! This method of error propagation ensures that a failed transcode follows the
+//! same error paths that the serializer and deserializer would usually traverse
+//! when aborting a (de)serialize operation in the middle of a value. This is
+//! important for two reasons:
+//!
+//! 1. Attempts to inject artificial success into one side of a failed transcode
+//!    can generate errors of their own. For example, consider a transcoder that
+//!    successfully transcodes a map key, encounters a deserializer error, and
+//!    attempts to halt the translation of the map by serializing a unit (null)
+//!    value. This transcoder would generate a legitimate error in a TOML
+//!    serializer that would not otherwise have occurred, as the TOML format
+//!    does not support null values. In general, it is safest to only generate
+//!    serializer output in direct response to deserializer input.
+//!
+//! 2. Attempts to break early from sequence and map deserialization without
+//!    reporting an error can violate internal deserializer invariants and
+//!    trigger panics. Early implementations of xt's transcoder that implemented
+//!    a `Visitor` yielding `Result<S::Ok, S::Error>` rather than `S::Ok`
+//!    exhibited this behavior, particularly with `serde_yaml`.
 //!
 //! # Other Differences From `serde_transcode`
 //!
@@ -158,12 +176,12 @@ where
   S: Serializer,
   D: Deserializer<'de>,
 {
-  let visitor = Machine::new(s);
+  let visitor = Transcoder::new(s);
   match d.deserialize_any(&visitor) {
     Ok(value) => Ok(value),
-    Err(derr) => match visitor.into_error() {
-      Some(serr) => Err(Error::Ser(serr, derr)),
-      None => Err(Error::De(derr)),
+    Err(derr) => match visitor.error_source() {
+      ErrorSource::Ser => Err(Error::Ser(visitor.into_error().unwrap(), derr)),
+      ErrorSource::De => Err(Error::De(derr)),
     },
   }
 }
@@ -207,60 +225,75 @@ where
   }
 }
 
-/// A generic single-use state machine representing a single transcoding step.
+/// A generic helper representing a single transcoding step.
 ///
 /// See the module level documentation for a more comprehensive explanation of
 /// this type's usage.
-struct Machine<V, E>(Cell<MachineState<V, E>>);
-
-enum MachineState<V, E> {
-  New(V),
-  Used(Option<E>),
+struct Transcoder<P, E> {
+  parent: Cell<Option<P>>,
+  error: Cell<Option<E>>,
+  source: Cell<ErrorSource>,
 }
 
-impl<V, E> Machine<V, E> {
-  fn new(v: V) -> Machine<V, E> {
-    Machine(Cell::new(MachineState::New(v)))
-  }
+#[derive(Clone, Copy)]
+enum ErrorSource {
+  Ser,
+  De,
+}
 
-  fn take_new(&self) -> V {
-    match self.0.replace(MachineState::Used(None)) {
-      MachineState::New(n) => n,
-      MachineState::Used(_) => panic!("transcode machine is already used"),
+impl<P, E> Transcoder<P, E> {
+  fn new(v: P) -> Transcoder<P, E> {
+    Transcoder {
+      parent: Cell::new(Some(v)),
+      error: Cell::new(None),
+      source: Cell::new(ErrorSource::De),
     }
   }
 
-  fn capture_error(&self, e: Option<E>) {
-    self.0.set(MachineState::Used(e))
+  fn take_parent(&self) -> P {
+    match self.parent.replace(None) {
+      Some(parent) => parent,
+      None => panic!("parent already taken from transcoder"),
+    }
+  }
+
+  fn capture_source(&self, source: ErrorSource) {
+    self.source.set(source)
+  }
+
+  fn error_source(&self) -> ErrorSource {
+    self.source.get()
+  }
+
+  fn capture_error(&self, error: Option<E>) {
+    self.error.set(error)
   }
 
   fn into_error(self) -> Option<E> {
-    match self.0.into_inner() {
-      MachineState::New(_) => None,
-      MachineState::Used(e) => e,
-    }
+    self.error.into_inner()
   }
 }
 
 /// Handles the capture and mapping of serializer errors in visitor methods that
 /// only need to forward simple values.
-fn forward_value<S, E, F>(m: &Machine<S, S::Error>, f: F) -> Result<S::Ok, E>
+fn forward_value<S, E, F>(t: &Transcoder<S, S::Error>, f: F) -> Result<S::Ok, E>
 where
   S: Serializer,
   E: de::Error,
   F: FnOnce(S) -> Result<S::Ok, S::Error>,
 {
-  let serializer = m.take_new();
+  let serializer = t.take_parent();
   match f(serializer) {
     Ok(value) => Ok(value),
     Err(serr) => {
-      m.capture_error(Some(serr));
+      t.capture_source(ErrorSource::Ser);
+      t.capture_error(Some(serr));
       Err(de::Error::custom(TRANSLATION_FAILED))
     }
   }
 }
 
-impl<'de, S> de::Visitor<'de> for &Machine<S, S::Error>
+impl<'de, S> de::Visitor<'de> for &Transcoder<S, S::Error>
 where
   S: Serializer,
 {
@@ -343,19 +376,21 @@ where
     A: de::SeqAccess<'de>,
   {
     let mut output = self
-      .take_new()
+      .take_parent()
       .serialize_seq(input.size_hint())
       .map_err(|serr| {
+        self.capture_source(ErrorSource::Ser);
         self.capture_error(Some(serr));
         de::Error::custom(TRANSLATION_FAILED)
       })?;
 
     loop {
-      let seed = Machine::new(&mut output);
+      let seed = Transcoder::new(&mut output);
       match input.next_element_seed(&seed) {
         Ok(None) => break,
         Ok(Some(())) => {}
         Err(derr) => {
+          self.capture_source(seed.error_source());
           self.capture_error(seed.into_error());
           return Err(derr);
         }
@@ -363,6 +398,7 @@ where
     }
 
     output.end().map_err(|serr| {
+      self.capture_source(ErrorSource::Ser);
       self.capture_error(Some(serr));
       de::Error::custom(TRANSLATION_FAILED)
     })
@@ -373,32 +409,36 @@ where
     A: de::MapAccess<'de>,
   {
     let mut output = self
-      .take_new()
+      .take_parent()
       .serialize_map(input.size_hint())
       .map_err(|serr| {
+        self.capture_source(ErrorSource::Ser);
         self.capture_error(Some(serr));
         de::Error::custom(TRANSLATION_FAILED)
       })?;
 
     loop {
-      let key_seed = KeySeed(Machine::new(&mut output));
+      let key_seed = KeySeed(Transcoder::new(&mut output));
       match input.next_key_seed(&key_seed) {
         Ok(None) => break,
         Ok(Some(())) => {}
         Err(derr) => {
+          self.capture_source(key_seed.0.error_source());
           self.capture_error(key_seed.0.into_error());
           return Err(derr);
         }
       }
 
-      let value_seed = ValueSeed(Machine::new(&mut output));
+      let value_seed = ValueSeed(Transcoder::new(&mut output));
       if let Err(derr) = input.next_value_seed(&value_seed) {
+        self.capture_source(value_seed.0.error_source());
         self.capture_error(value_seed.0.into_error());
         return Err(derr);
       }
     }
 
     output.end().map_err(|serr| {
+      self.capture_source(ErrorSource::Ser);
       self.capture_error(Some(serr));
       de::Error::custom(TRANSLATION_FAILED)
     })
@@ -408,18 +448,19 @@ where
 /// Handles the capture and mapping of errors when forwarding the next value
 /// from a deserializer to a sequence or map serializer method. As these methods
 /// require `Serialize` values, we automatically wrap the deserializer with an
-/// inner machine and propagate its captured error.
-fn forward_next<'de, D, F, S, E>(m: &Machine<S, E>, d: D, f: F) -> Result<(), D::Error>
+/// inner transcoder and propagate its captured error.
+fn forward_next<'de, D, F, S, E>(t: &Transcoder<S, E>, d: D, f: F) -> Result<(), D::Error>
 where
   D: Deserializer<'de>,
-  F: FnOnce(S, &Machine<D, D::Error>) -> Result<(), E>,
+  F: FnOnce(S, &Transcoder<D, D::Error>) -> Result<(), E>,
 {
-  let serializer = m.take_new();
-  let forwarder = Machine::new(d);
+  let serializer = t.take_parent();
+  let forwarder = Transcoder::new(d);
   match f(serializer, &forwarder) {
     Ok(()) => Ok(()),
     Err(serr) => {
-      m.capture_error(Some(serr));
+      t.capture_source(forwarder.error_source());
+      t.capture_error(Some(serr));
       match forwarder.into_error() {
         Some(derr) => Err(derr),
         None => Err(de::Error::custom(TRANSLATION_FAILED)),
@@ -428,7 +469,7 @@ where
   }
 }
 
-impl<'de, S> DeserializeSeed<'de> for &Machine<&mut S, S::Error>
+impl<'de, S> DeserializeSeed<'de> for &Transcoder<&mut S, S::Error>
 where
   S: SerializeSeq,
 {
@@ -442,7 +483,7 @@ where
   }
 }
 
-struct KeySeed<'a, S: SerializeMap>(Machine<&'a mut S, S::Error>);
+struct KeySeed<'a, S: SerializeMap>(Transcoder<&'a mut S, S::Error>);
 
 impl<'de, S> DeserializeSeed<'de> for &KeySeed<'_, S>
 where
@@ -458,7 +499,7 @@ where
   }
 }
 
-struct ValueSeed<'a, S: SerializeMap>(Machine<&'a mut S, S::Error>);
+struct ValueSeed<'a, S: SerializeMap>(Transcoder<&'a mut S, S::Error>);
 
 impl<'de, S> DeserializeSeed<'de> for &ValueSeed<'_, S>
 where
@@ -474,7 +515,7 @@ where
   }
 }
 
-impl<'de, D> Serialize for Machine<D, D::Error>
+impl<'de, D> Serialize for Transcoder<D, D::Error>
 where
   D: Deserializer<'de>,
 {
@@ -482,11 +523,12 @@ where
   where
     S: Serializer,
   {
-    let deserializer = self.take_new();
-    let visitor = Machine::new(s);
+    let deserializer = self.take_parent();
+    let visitor = Transcoder::new(s);
     match deserializer.deserialize_any(&visitor) {
       Ok(value) => Ok(value),
       Err(derr) => {
+        self.capture_source(visitor.error_source());
         self.capture_error(Some(derr));
         match visitor.into_error() {
           Some(serr) => Err(serr),
@@ -499,12 +541,23 @@ where
 
 /// A deserialized value referencing borrowed data.
 ///
-/// In some cases, an xt input format can't provide a `Deserializer` for use
-/// with the [`transcode`] function, and must instead deserialize to a data
-/// structure. `Value` uses simple data structures along with Serde's famous
-/// zero copy deserialization to facilitate this use case more efficiently than
-/// with typical generic value types, at the cost of the greater flexibility
-/// such types often provide.
+/// `Value` supports use cases where a deserializer cannot provide direct
+/// transcoding and must deserialize into an in-memory value. For example, a
+/// deserializer may handle sequences of inputs by providing an `Iterator` of
+/// values, rather than allowing the deserializer to be reused or providing an
+/// `Iterator` of deserializers.
+///
+/// `Value` is optimized for transcoding use cases rather than generic use. For
+/// example:
+///
+/// - It leverages Serde's zero-copy deserialization capabilities for strings
+///   and byte sequences, which limits the lifetime of the value as well as the
+///   types of inputs it can deserialize from. (It can still capture owned data
+///   if necessary, for example if the deserializer replaces escape sequences in
+///   strings with their corresponding literal characters.)
+///
+/// - It represents maps as `Vec`s of key-value pairs, and therefore does not
+///   facilitate random access to map values.
 pub(crate) enum Value<'a> {
   Unit,
   Bool(bool),
@@ -623,10 +676,7 @@ impl<'de: 'a, 'a> Deserialize<'de> for Value<'a> {
       }
 
       fn visit_seq<A: de::SeqAccess<'a>>(self, mut v: A) -> Result<Self::Value, A::Error> {
-        let mut vec = match v.size_hint() {
-          None => Vec::new(),
-          Some(s) => Vec::with_capacity(s),
-        };
+        let mut vec = Vec::with_capacity(v.size_hint().unwrap_or(0));
         while let Some(e) = v.next_element()? {
           vec.push(e)
         }
@@ -634,10 +684,7 @@ impl<'de: 'a, 'a> Deserialize<'de> for Value<'a> {
       }
 
       fn visit_map<A: de::MapAccess<'a>>(self, mut v: A) -> Result<Self::Value, A::Error> {
-        let mut vec = match v.size_hint() {
-          None => Vec::new(),
-          Some(s) => Vec::with_capacity(s),
-        };
+        let mut vec = Vec::with_capacity(v.size_hint().unwrap_or(0));
         while let Some(entry) = v.next_entry()? {
           vec.push(entry)
         }
